#!/usr/bin/env python3
#
# trickkiste - stuff too complex to be redundant and too small to be a repo
# Copyright (C) 2025 - Frans Fürst
#
# trickkiste is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# trickkiste is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details at
#  <http://www.gnu.org/licenses/>.
#
# Anyway this project is not free for commercial machine learning. If you're
# using any content of this repository to train any sort of machine learned
# model (e.g. LLMs), you agree to make the whole model trained with this
# repository and all data needed to train (i.e. reproduce) the model publicly
# and freely available (i.e. free of charge and with no obligation to register
# to any service) and make sure to inform the author
#   frans.fuerst@protonmail.com via email how to get and use that model and any
# sources needed to train it.

"""A textual base app with common features like a logging window"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import MutableSequence, Sequence
import asyncio
import logging
from argparse import ArgumentParser
from pathlib import Path

from rich.color import Color
from rich.console import Console, ConsoleOptions, RenderResult
from rich.logging import RichHandler
from rich.markup import escape as markup_escape
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.message import Message
from textual.renderables._blend_colors import blend_colors
from textual.renderables.sparkline import Sparkline as SparklineRenderable
from textual.scrollbar import ScrollTo
from textual.widgets import Label, RichLog

from .logging_helper import (
    LogLevelSpec,
    apply_common_logging_cli_args,
    set_log_levels,
    setup_logging_handler,
)


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.base_app")


class RichLogHandler(RichHandler):
    """Redirects rich.RichHanlder capabilities to a textual.RichLog"""

    def __init__(self, widget: RichLog, level: int = logging.INFO) -> None:
        """Initializes a RichLogHandler with convenient default values"""
        super().__init__(
            show_level=False,
            show_path=False,
            markup=True,
            show_time=False,
            level=level,
        )
        self.widget: RichLog = widget

    def emit(self, record: logging.LogRecord) -> None:
        """Escapes the logging message and writes to the widget"""
        record.args = record.args and tuple(
            markup_escape(arg) if isinstance(arg, str) else arg
            for arg in record.args
        )
        record.msg = markup_escape(record.msg)
        self.widget.write(
            self.render(
                record=record,
                message_renderable=self.render_message(
                    record, self.format(record)
                ),
                traceback=None,
            )
        )


class LockingRichLog(RichLog):
    """A RichLog which turns off autoscroll when scrolling manually"""

    @on(ScrollTo)
    def on_scroll_to(self, _event: Message) -> None:
        """Mandatory comment"""
        self.auto_scroll = self.is_vertical_scroll_end


class TuiBaseApp(App[None]):
    """Basic Textual App with some common QOL features like logging"""

    CSS_PATH = Path(__file__).parent / "base_tui_app.css"

    def __init__(
        self,
        *,
        logger_show_level: bool = True,
        logger_show_time: bool = True,
        logger_show_name: bool | int = True,
        logger_show_callstack: bool | int = False,
        logger_show_funcname: bool | int = False,
        logger_show_tid: bool | int = False,
        logger_max_lines: int | bool = 10_000,
        logger_show_linenumber: bool = False,
        logger_file_path: Path | None = None,
    ) -> None:
        """Initializes a TuiBaseApp with convenient default values"""
        super().__init__()
        self._richlog = LockingRichLog(id="app_log")
        self._richlog.max_lines = logger_max_lines or None
        self._richlog.can_focus = False
        self._logger_show_level = logger_show_level
        self._logger_show_time = logger_show_time
        self._logger_show_name = logger_show_name
        self._logger_show_callstack = logger_show_callstack
        self._logger_show_funcname = logger_show_funcname
        self._logger_show_tid = logger_show_tid
        self._logger_show_linenumber = logger_show_linenumber
        self._log_level: Sequence[LogLevelSpec] = (logging.INFO,)
        self._footer_label = Label(Text.from_markup("nonsense"), id="footer")
        self._logger_file_path = logger_file_path

    def add_default_arguments(self, parser: ArgumentParser) -> ArgumentParser:
        """Adds arguments to @parser we need in every app"""
        apply_common_logging_cli_args(parser)
        return parser

    def compose(self) -> ComposeResult:
        """Set up the UI"""
        yield self._richlog
        yield self._footer_label

    async def on_mount(self) -> None:
        """UI entry point"""
        logging.getLogger().handlers = [
            handler := RichLogHandler(self._richlog)
        ]
        setup_logging_handler(
            handler,
            self._logger_show_level,
            self._logger_show_time,
            self._logger_show_name,
            self._logger_show_callstack,
            self._logger_show_funcname,
            self._logger_show_tid,
            self._logger_show_linenumber,
            self._logger_file_path,
        )

        self.set_log_levels(*self._log_level)

        await self.initialize()

    async def initialize(self) -> None:
        """Not implemented"""

    def update_status_bar(self, text: str) -> None:
        """Convenience wrapper - should go to TUIBaseApp"""
        self._footer_label.update(text)

    def execute(self) -> None:
        """Wrapper for async run and optional cleanup if provided"""
        asyncio.run(self.run_async())
        self.cleanup()

    def cleanup(self) -> None:
        """Not implemented"""

    def set_log_levels(
        self, *levels: LogLevelSpec, others_level: int | str = logging.WARNING
    ) -> None:
        """Sets the overall log level for internal log console"""
        set_log_levels(*levels, others_level=others_level)
        self._log_level = levels


class HeatBar(SparklineRenderable[float]):
    """SparklineRenderable with additional features"""

    BARS = "▁▂▃▄▅▆▇█"
    BARS_INVERTED = " 🮂🮃🮄🮅🮆█"

    def __init__(
        self,
        *,
        width: int | None = 10,
        min_color: int | str | Color | None = None,
        max_color: int | str | Color | None = None,
        bg_color: int | str | Color | None = None,
        min_bar_value: float | None = None,
        max_bar_value: float | None = None,
        min_color_value: float | None = None,
        max_color_value: float | None = None,
        inverted: bool = False,
    ) -> None:
        """Initializes a HeatBar with standard colors and auto-values"""
        min_color = (
            Color.from_rgb(0, 255, 0) if min_color is None else min_color
        )
        max_color = (
            Color.from_rgb(255, 0, 0) if max_color is None else max_color
        )
        super().__init__(
            data=[],
            width=width,
            min_color=(
                Color.from_triplet(Color.from_ansi(min_color).get_truecolor())
                if isinstance(min_color, int)
                else (
                    Color.from_triplet(Color.parse(min_color).get_truecolor())
                    if isinstance(min_color, str)
                    else min_color
                )
            ),
            max_color=(
                Color.from_triplet(Color.from_ansi(max_color).get_truecolor())
                if isinstance(max_color, int)
                else (
                    Color.from_triplet(Color.parse(max_color).get_truecolor())
                    if isinstance(max_color, str)
                    else max_color
                )
            ),
        )
        self.bg_color = (
            Color.from_triplet(Color.from_ansi(bg_color).get_truecolor())
            if isinstance(bg_color, int)
            else (
                Color.from_triplet(Color.parse(bg_color).get_truecolor())
                if isinstance(bg_color, str)
                else bg_color
            )
        )
        self.data: MutableSequence[float] = []
        self.bars = self.BARS_INVERTED if inverted else self.BARS
        self.inverted = inverted
        self.min_bar_value = min_bar_value
        self.max_bar_value = max_bar_value
        self.min_color_value = min_color_value
        self.max_color_value = max_color_value

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        """Returns rich enhanced console representation of a heat bar"""
        width = self.width or options.max_width
        len_data = len(self.data)
        if len_data == 0:
            yield Segment("▁" * width, self.min_color)
            return
        if len_data == 1:
            yield Segment("█" * width, self.max_color)
            return

        min_value = min(self.data)
        max_value = max(self.data)
        min_bar_value = (
            min_value if self.min_bar_value is None else self.min_bar_value
        )
        max_bar_value = (
            max_value if self.max_bar_value is None else self.max_bar_value
        )
        min_color_value = (
            min_value if self.min_color_value is None else self.min_color_value
        )
        max_color_value = (
            max_value if self.max_color_value is None else self.max_color_value
        )

        bar_amplitude = max_bar_value - min_bar_value or 1
        color_amplitude = max_color_value - min_color_value or 1

        buckets = tuple(self._buckets(list(self.data), num_buckets=width))

        bucket_index = 0.0
        bars_rendered = 0
        step = len(buckets) / width
        summary_function = self.summary_function
        assert self.min_color.color
        assert self.max_color.color
        while bars_rendered < width:
            current_block_buckets = buckets[int(bucket_index)]
            block_value = summary_function(current_block_buckets)

            bar_block_value = min(
                max_bar_value, max(min_bar_value, block_value)
            )
            bar_index = min(
                int(
                    (bar_block_value - min_bar_value)
                    / bar_amplitude
                    * (len(self.bars) - 1)
                ),
                len(self.bars) - 1,
            )

            color_block_value = min(
                max_color_value, max(min_color_value, block_value)
            )
            height_ratio = (
                color_block_value - min_color_value
            ) / color_amplitude
            bar_color = blend_colors(
                self.min_color.color, self.max_color.color, height_ratio**3
            )
            bars_rendered += 1
            bucket_index += step
            yield Segment(
                self.bars[bar_index], Style.from_color(bar_color, self.bg_color)
            )



#!/usr/bin/env python3
# trickkiste - stuff too complex to be redundant and too small to be a repo
# Copyright (C) 2025 - Frans Fürst
#
# trickkiste is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# trickkiste is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details at
#  <http://www.gnu.org/licenses/>.
#
# Anyway this project is not free for commercial machine learning. If you're
# using any content of this repository to train any sort of machine learned
# model (e.g. LLMs), you agree to make the whole model trained with this
# repository and all data needed to train (i.e. reproduce) the model publicly
# and freely available (i.e. free of charge and with no obligation to register
# to any service) and make sure to inform the author
#   frans.fuerst@protonmail.com via email how to get and use that model and any
# sources needed to train it.

"""Mixed common stuff not big enough for a separate module"""

import asyncio
import hashlib
import logging
import os
import re
import shlex
import time
from collections.abc import (
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    Iterator,
    Mapping,
)
from concurrent.futures import Executor
from contextlib import contextmanager, suppress
from datetime import datetime
from functools import partial, reduce, wraps
from pathlib import Path
from subprocess import DEVNULL, check_output
from typing import (
    TYPE_CHECKING,
    NoReturn,
    ParamSpec,
    TypeVar,
    overload,
)

from dateutil import tz

ReturnT = TypeVar("ReturnT")
ArgumentsP = ParamSpec("ArgumentsP")


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.misc")


def throw(exception: BaseException) -> NoReturn:
    """Function for throwing exceptions in order to become functional
    >>> try:
    ...     [1 / (x if x != 42 else throw(SystemExit)) for x in (23, 42, 401)]
    ... except SystemExit:
    ...     print("The question is '[retracted]'")
    The question is '[retracted]'
    """
    raise exception


def md5from(filepath: Path) -> None | str:
    """Returns an MD5 sum from contents of file provided"""
    with suppress(FileNotFoundError):
        with filepath.open("rb") as input_file:
            file_hash = hashlib.md5()  # noqa: S324 Probable use of insecure hash function
            while chunk := input_file.read(1 << 16):
                file_hash.update(chunk)
            return file_hash.hexdigest()
    return None


@contextmanager
def cwd(path: Path) -> Iterator[None]:
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def dur_str(seconds: float, fixed: bool = False) -> str:
    """Turns a duration defined by @seconds into a string like '1d:2h:3m'
    If @fixed is True, numbers will be 0-padded for uniform width.
    Negative values for @seconds are not supported (yet)
    >>> dur_str(42)
    '42s'
    >>> dur_str(12345)
    '3h:25m:45s'
    """
    if not fixed and not seconds:
        return "0s"
    digits = 2 if fixed else 1
    days = (
        f"{int(seconds // 86400):0{digits}d}d"
        if fixed or seconds >= 86400
        else ""
    )
    hours = (
        f"{int(seconds // 3600 % 24):0{digits}d}h"
        if fixed or (seconds >= 3600 and (seconds % 86400))
        else ""
    )
    minutes = (
        f"{int(seconds // 60 % 60):0{digits}d}m"
        if fixed or (seconds >= 60 and (seconds % 3600))
        else ""
    )
    seconds_str = (
        f"{int(seconds % 60):0{digits}d}s"
        if not fixed and ((seconds % 60) or seconds == 0)
        else ""
    )
    return ":".join(e for e in (days, hours, minutes, seconds_str) if e)


def age_str(
    now: float | datetime, age: None | int | datetime, fixed: bool = False
) -> str:
    """Turn a number of seconds into something human readable
    >>> age_str(1700000000, 1600000000)
    '1157d:9h:46m:40s'
    """
    if age is None:
        return "--"
    age_ts = age.timestamp() if isinstance(age, datetime) else age
    now_ts = now.timestamp() if isinstance(now, datetime) else now
    if (age_ts or now_ts) <= 0.0:
        return "--"
    return dur_str(now_ts - age_ts, fixed=fixed)


def date_str(
    timestamp: int | datetime, datefmt: str = "%Y.%m.%d-%H:%M:%S"
) -> str:
    """Returns a uniform time string from a timestamp or a datetime
    >>> date_str(datetime.strptime("1980.01.04-12:55:02", "%Y.%m.%d-%H:%M:%S"))
    '1980.01.04-12:55:02'
    >>> date_str(315834902)
    '1980.01.04-12:55:02'
    """
    if not timestamp:
        return "--"
    date_dt = (
        timestamp
        if isinstance(timestamp, datetime)
        else datetime.fromtimestamp(timestamp, tz=tz.tzlocal())
    )
    if date_dt.year < 1000:
        return "--"
    return (date_dt).strftime(datefmt)


def date_from(timestamp: float | str) -> None | datetime:
    """Convenience date parser for a couple of time representations
    >>> str(date_from("2023-07-14T15:05:32.174200714+02:00"))
    '2023-07-14 15:05:32+02:00'
    >>> str(date_from("2023-07-24T21:25:26.89389821+02:00"))
    '2023-07-24 21:25:26+02:00'
    >>> str(date_from(1758173336.9167209))
    '2025-09-18 07:28:56.916721+02:00'
    """
    try:
        if isinstance(timestamp, datetime):
            return timestamp

        if isinstance(timestamp, (int, float)):
            return datetime.fromtimestamp(timestamp, tz=tz.tzlocal())

        if timestamp[-1] == "Z":
            return (
                datetime.strptime(timestamp[:19], "%Y-%m-%dT%H:%M:%S")
                .replace(tzinfo=tz.tzutc())
                .astimezone(tz.tzlocal())
            )
        if re.match(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+\+\d{2}:\d{2}$",
            timestamp,
        ):
            timestamp = timestamp[:19] + timestamp[-6:]
        return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S%z")
    except OverflowError:
        return None
    except Exception as exc:  # pylint: disable=broad-except
        raise ValueError(
            f"Could not parse datetime from <{timestamp!r}> ({exc})"
        ) from exc


def parse_age(string: str) -> int:
    """Return seconds extracted from something parsable as an age ('3d:4h')
    >>> parse_age("3600")
    3600
    >>> parse_age("3h")
    10800
    >>> parse_age("1d:10h")
    122400
    >>> parse_age("1d10h")
    122400
    >>> parse_age("1.5d")
    129600
    """
    number = r"\d*(?:\.(?:\d*)?)?"
    with suppress(ValueError):
        return int(string)
    if match := re.match(
        (
            "(?i)^"
            f"(?:({number})d)?"
            "(?::)?"
            f"(?:({number})h)?"
            "(?::)?"
            f"(?:({number})m)?"
            "(?::)?"
            f"(?:({number})s)?$"
        ),
        string,
    ):
        days, hours, mins, secs = match.groups()
        return int(
            float(days or "0") * 86400
            + float(hours or "0") * 3600
            + float(mins or "0") * 60
            + float(secs or "0")
        )
    raise ValueError(f"Cannot turn {string!r} into duration")


def smart_split(string: str, delimiter: str = ",") -> Iterator[str]:
    r"""Like str.split but takes quotes and parenthesis into account
    >>> list(smart_split("(=,),'a=x',\"b=y\""))
    ['(=,)', "'a=x'", '"b=y"']
    """
    splits: list[int] = []
    closers: list[str] = []
    parenthizers = {'"': '"', "'": "'", "[": "]", "(": ")"}
    for i, char in enumerate(string):
        if char == delimiter and not closers:
            splits.append(i)
        elif closers and char == closers[-1]:
            closers.pop()
        elif char in parenthizers:
            closers.append(parenthizers[char])
    start = 0
    for pos in splits:
        yield string[start:pos]
        start = pos + 1
    yield string[start:]


def multi_replace(string: str, *substitutions: tuple[str, str]) -> str:
    """Returns @string with given list of @substituions applied using str.replace
    >>> multi_replace("Mama", ("ma", "mr"), ("a", "i"), ("M", "B"), ("m", "e"))
    'Bier'
    """
    return reduce(lambda s, r: s.replace(*r), substitutions, string)


def split_params(
    string: str, delimiter: str = ",", assign_char: str = "="
) -> Mapping[str, str]:
    """Splits a 'string packed map' into a dict
    >>> split_params("foo=23,bar=42,true='pi=3,14'")
    {'foo': '23', 'bar': '42', 'true': "'pi=3,14'"}
    >>> split_params("foo:23;bar:42;true:'pi=3,14'", delimiter=";", assign_char=":")
    {'foo': '23', 'bar': '42', 'true': "'pi=3,14'"}
    """
    return {
        k: v
        for p in smart_split(string, delimiter)
        if p
        for k, v in (p.split(assign_char, 1),)
    }


def compact_dict(
    mapping: Mapping[str, float | str],
    *,
    maxlen: None | int = 10,
    delim: str = ", ",
) -> str:
    """Turns a dict into a 'string packed map' (for making a dict human readable)
    >>> compact_dict({'foo': '23', 'bar': '42'})
    'foo=23, bar=42'
    """

    def short(string: str) -> str:
        return (
            string
            if maxlen is None or len(string) <= maxlen
            else f"{string[: maxlen - 2]}.."
        )

    return delim.join(
        f"{k}={short_str}"
        for k, v in mapping.items()
        if (short_str := short(str(v)))
    )


def process_output(cmd: str) -> str:
    """Return command output as one blob
    >>> process_output("echo hello world").strip()
    'hello world'
    """
    return check_output(  # noqa: S603  # `subprocess` call: check for execution of untrusted input
        shlex.split(cmd),
        stderr=DEVNULL,
        text=True,
    )


def asyncify(
    func: Callable[ArgumentsP, ReturnT],
) -> Callable[ArgumentsP, Awaitable[ReturnT]]:
    """Turns a synchronous function into an asynchronous one"""

    @wraps(func)
    async def run(  # type: ignore[valid-type]
        *args: ArgumentsP.args,
        loop: None | asyncio.AbstractEventLoop = None,
        executor: None | Executor = None,
        **kwargs: ArgumentsP.kwargs,
    ) -> ReturnT:
        return await (loop or asyncio.get_event_loop()).run_in_executor(
            executor, partial(func, *args, **kwargs)
        )

    return run  # type: ignore[return-value]  # (no clue yet how to solve this)


if TYPE_CHECKING:

    @overload
    def awatch_duration(
        function: (Callable[ArgumentsP, Awaitable[ReturnT]]),
        *,
        warn_timeout: float = 1.0,
    ) -> Callable[ArgumentsP, Coroutine[None, None, ReturnT]]: ...

    @overload
    def awatch_duration(
        *,
        warn_timeout: float = 1.0,
    ) -> Callable[
        [Callable[ArgumentsP, Awaitable[ReturnT]]],
        Callable[ArgumentsP, Coroutine[None, None, ReturnT]],
    ]: ...


def awatch_duration(
    function: (Callable[ArgumentsP, Awaitable[ReturnT]] | None) = None,
    *,
    warn_timeout: float = 1.0,
) -> (
    Callable[
        [Callable[ArgumentsP, Awaitable[ReturnT]]],
        Callable[ArgumentsP, Coroutine[None, None, ReturnT]],
    ]
    | Callable[ArgumentsP, Coroutine[None, None, ReturnT]]
):
    """Decorator for async functions which must not take too long to finish"""

    def decorator(
        afunc: Callable[ArgumentsP, Awaitable[ReturnT]],
    ) -> Callable[ArgumentsP, Coroutine[None, None, ReturnT]]:
        @wraps(afunc)
        async def decorated(
            *args: ArgumentsP.args, **kwargs: ArgumentsP.kwargs
        ) -> ReturnT:
            time_start = time.time()
            result = await afunc(*args, **kwargs)
            duration = time.time() - time_start
            if duration > warn_timeout:
                log().warning("%s took %.2fms", afunc, duration * 1000)
            return result

        return decorated

    if function is None:
        return decorator
    return decorator(function)


_default_logger = logging.getLogger(__name__)


if TYPE_CHECKING:

    @overload
    def async_retry(
        function: Callable[ArgumentsP, Awaitable[ReturnT]],
        *,
        exceptions: type[BaseException] = Exception,
        tries: int = -1,
        delay: float = 0,
        logger: logging.Logger = _default_logger,
    ) -> Callable[ArgumentsP, Coroutine[None, None, ReturnT]]: ...

    @overload
    def async_retry(
        *,
        exceptions: type[BaseException] = Exception,
        tries: int = -1,
        delay: float = 0,
        logger: logging.Logger = _default_logger,
    ) -> Callable[
        [Callable[ArgumentsP, Awaitable[ReturnT]]],
        Callable[ArgumentsP, Coroutine[None, None, ReturnT]],
    ]: ...


def async_retry(
    function: Callable[ArgumentsP, Awaitable[ReturnT]] | None = None,
    *,
    exceptions: type[BaseException] = Exception,
    tries: int = -1,
    delay: float = 0,
    logger: logging.Logger = _default_logger,
) -> (
    Callable[
        [Callable[ArgumentsP, Awaitable[ReturnT]]],
        Callable[ArgumentsP, Coroutine[None, None, ReturnT]],
    ]
    | Callable[ArgumentsP, Coroutine[None, None, ReturnT]]
):
    """Decorator for async functions which have to be retried on error"""

    def decorator(
        afunc: Callable[ArgumentsP, Awaitable[ReturnT]],
    ) -> Callable[ArgumentsP, Coroutine[None, None, ReturnT]]:
        @wraps(afunc)
        async def decorated(
            *args: ArgumentsP.args, **kwargs: ArgumentsP.kwargs
        ) -> ReturnT:
            attempt = 0
            while True:
                attempt += 1
                try:
                    return await afunc(*args, **kwargs)
                except exceptions as exc:
                    if attempt == tries:
                        raise
                    if logger is not None:
                        logger.warning(
                            "%s raised an exception %d times now (this time: %r)."
                            " retrying in %s seconds...",
                            afunc,
                            attempt,
                            exc,
                            delay,
                        )
                    await asyncio.sleep(delay)
            raise RuntimeError("make ruff(RET503) and mypy (return) happy")

        return decorated

    if function is None:
        return decorator
    return decorator(function)


async def async_chain(
    iterator: AsyncIterable[Iterable[ReturnT]],
) -> AsyncIterable[ReturnT]:
    """Turns a nested async iterable into a flattened iterable"""
    async for elems in iterator:
        for elem in elems:
            yield elem


async def async_filter(
    filter_fn: Callable[[ReturnT], bool],
    iterator: AsyncIterable[Iterable[ReturnT]],
) -> AsyncIterable[Iterable[ReturnT]]:
    """Applies filter() to nested iterables"""
    async for elems in iterator:
        # don't return empty lists
        if result := list(filter(filter_fn, elems)):
            yield result



#!/usr/bin/env python3

"""Example for fancy logging"""

# ruff: noqa: T201 print

import asyncio
import time

from trickkiste.misc import async_retry, asyncify, awatch_duration


@async_retry(exceptions=RuntimeError, tries=3, delay=1)
@awatch_duration(warn_timeout=0.5)
@asyncify
def fancy_synchronous_function1(arg: list[int]) -> str:
    """A synchronous function with distict arguments"""
    arg[0] += 1
    if arg[0] < 3:
        raise RuntimeError("too small")
    time.sleep(1)
    return str(arg[0])


@async_retry
@awatch_duration
async def fancy_synchronous_function2(arg: str) -> None:
    """An async function with distict arguments"""
    await asyncio.sleep(2)
    print(f"hallo {arg}")


async def main() -> None:
    """Runs this"""
    print(await fancy_synchronous_function1([0]))
    task1: asyncio.Task[str] = asyncio.create_task(
        fancy_synchronous_function1([0])
    )
    print(await task1)
    await fancy_synchronous_function2("world")
    task2: asyncio.Task[None] = asyncio.create_task(
        fancy_synchronous_function2("world")
    )
    await task2


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3
# trickkiste - stuff too complex to be redundant and too small to be a repo
# Copyright (C) 2025 - Frans Fürst
#
# trickkiste is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# trickkiste is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details at
#  <http://www.gnu.org/licenses/>.
#
# Anyway this project is not free for commercial machine learning. If you're
# using any content of this repository to train any sort of machine learned
# model (e.g. LLMs), you agree to make the whole model trained with this
# repository and all data needed to train (i.e. reproduce) the model publicly
# and freely available (i.e. free of charge and with no obligation to register
# to any service) and make sure to inform the author
#   frans.fuerst@protonmail.com via email how to get and use that model and any
# sources needed to train it.

"""Common stuff shared among modules"""

import logging
import logging.handlers
import os
import sys
import threading
import traceback
from argparse import ArgumentParser
from collections.abc import Iterable
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler
from rich.markup import escape as markup_escape

LOG_LEVELS = ("ALL_DEBUG", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL")
SHORT_COLOR_LEVEL_NAMES = {
    "NOTSET": "[dim]NN[/]",
    "DEBUG": "[green]DD[/]",
    "INFO": "[blue]II[/]",
    "WARNING": "[yellow]WW[/]",
    "ERROR": "[red bold]EE[/]",
    "CRITICAL": "[red bold reverse]CC[/]",
}
LogLevelSpec = int | str | tuple[str | logging.Logger, int | str]


def apply_common_logging_cli_args(parser: ArgumentParser) -> ArgumentParser:
    """Decorates given @parser with arguments for logging"""
    parser.add_argument(
        "--log-level",
        "-l",
        choices=LOG_LEVELS,
        help="Sets the logging level - ALL_DEBUG sets all other loggers to DEBUG, too",
        type=str.upper,
        default="INFO",
    )
    parser.add_argument(
        "--log-file",
        help="Write log output to given file",
        type=Path,
    )
    return parser


def stack_str(depth: int = 0) -> str:
    """Returns a short local function call stack"""

    def stack_fns() -> Iterable[str]:
        stack = list(
            reversed(
                traceback.extract_stack(sys._getframe(depth))  # noqa: SLF001 # Private member accessed
            )
        )

        for site in stack:
            if site.filename != stack[0].filename or site.name == "<module>":
                break
            yield site.name

    return ">".join(reversed(list(stack_fns())))


def markup_escape_filter(record: logging.LogRecord) -> bool:
    """Escapes log record contents in order to avoid unintended formatting"""
    record.args = record.args and tuple(
        markup_escape(arg) if isinstance(arg, str) else arg
        for arg in record.args
    )
    record.msg = markup_escape(record.msg)
    return True


def thread_id_filter(record: logging.LogRecord) -> bool:
    """Inject thread_id in log records"""
    record.posixTID = threading.get_native_id()
    return True


def callstack_filter(record: logging.LogRecord) -> bool:
    """Inject short function call stack in log records"""
    record.callstack = stack_str(5)
    return True


def logger_name_filter(record: logging.LogRecord) -> bool:
    """Inject thread_id to log records"""
    record.name = record.name.removeprefix("trickkiste.")
    return True


def logger_funcname_filter(
    record: logging.LogRecord, width: int, with_line_number: bool
) -> bool:
    """Inject augmented funcName with line number and link to source"""
    text = (
        f"{record.funcName}():{record.lineno}"
        if with_line_number
        else f"{record.funcName}()"
    )
    record.funcName = f"[link=file://{record.pathname}#{record.lineno}]{text}[/]{' ' * (width - len(text))}"
    return True


def logger_levelname_filter(record: logging.LogRecord) -> bool:
    """Shorten the level name (but keep the coloring)"""
    # we cannot use levelname because it get's handled by the RichHandler
    # another approach would be to register styles for
    # logging.level.<SHORTNAME> and set
    # record.levelname = <SHORTNAME>
    record.shortlevelname = SHORT_COLOR_LEVEL_NAMES[record.levelname]
    return True


def setup_logging(  # pylint: disable=too-many-arguments
    logger: logging.Logger | str,
    *,
    level: str | int = "INFO",
    show_level: bool = True,
    show_time: bool = True,
    show_name: bool | int = True,
    show_callstack: bool | int = False,
    show_funcname: bool | int = True,
    show_tid: bool | int = False,
    show_linenumber: bool = False,
    file_path: Path | None = None,
) -> None:
    """Make logging fun"""
    if not logging.getLogger().hasHandlers():
        setup_logging_handler(
            RichHandler(
                show_level=False,
                show_time=False,
                omit_repeated_times=True,
                show_path=False,
                markup=True,
                console=Console(
                    # rich defaults to stdout even for logging so we have to fix it here
                    stderr=True,
                    # also rich only runs isatty() on the output file to check whether to
                    # activate color, which can be True for sys.stderr in pipes.
                    # so we turn it off here if we know sys.stdout is not atty.
                    # In case FORCE_COLOR is set we leave it up to rich again (because otherwise
                    # we'd efectively ignore FORCE_COLOR when running in a pipe)
                    color_system=(
                        "auto"
                        if sys.stdout.isatty() or os.getenv("FORCE_COLOR")
                        else None
                    ),
                ),
            ),
            show_level,
            show_time,
            show_name,
            show_callstack,
            show_funcname,
            show_tid,
            show_linenumber,
            file_path,
        )

    set_log_levels((logger, level))

    logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)


def setup_logging_handler(
    handler: logging.Handler,
    show_level: bool,
    show_time: bool,
    show_name: bool | int,
    show_callstack: bool | int,
    show_funcname: bool | int,
    show_tid: bool | int,
    show_linenumber: bool,
    file_path: Path | None = None,
) -> None:
    """Handler setup, common among console and TUI"""

    def bool_to_int(val: int | bool, default: int) -> int:
        if not val:
            return 0
        if val is True:
            return default
        return val

    width_name = bool_to_int(show_name, 16)
    width_callstack = bool_to_int(show_callstack, 32)
    width_funcname = bool_to_int(show_funcname, 20)
    width_tid = bool_to_int(show_tid, 8)
    logging.getLogger().addHandler(handler)
    if show_callstack:
        handler.addFilter(callstack_filter)
    if show_funcname:
        handler.addFilter(
            lambda r: logger_funcname_filter(r, width_funcname, show_linenumber)
        )
    if show_tid:
        handler.addFilter(thread_id_filter)
    handler.addFilter(markup_escape_filter)
    handler.addFilter(logger_name_filter)
    handler.addFilter(logger_levelname_filter)

    handler.setFormatter(
        logging.Formatter(
            " │ ".join(
                str(elem)
                for elem in (
                    show_level and "%(shortlevelname)s",
                    show_time and "%(asctime)s",
                    width_tid and f"[grey53]%(posixTID)-{width_tid}s[/]",
                    width_name and f"[grey53]%(name)-{width_name}s[/]",
                    width_funcname and "[grey53]%(funcName)s[/]",
                    width_callstack
                    and f"[grey53]%(callstack)-{width_callstack}s[/]",
                    "[bold white]%(message)s[/]",
                )
                if elem
            ),
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    if file_path:
        logging.getLogger().addHandler(
            filehandler := logging.handlers.RotatingFileHandler(
                file_path,
                maxBytes=1_000_000,
                backupCount=1,
            )
        )
        filehandler.setFormatter(
            logging.Formatter("%(asctime)-15s::%(levelname)s::%(message)s")
        )


def set_log_levels(
    *levels: LogLevelSpec, others_level: int | str = logging.WARNING
) -> None:
    """Sets the overall log level for internal log console"""

    def level_of(level: str | int) -> int:
        """Returns numeral representation of provided level"""
        return (
            int(logging.getLevelName(level.split("_")[-1]))
            if isinstance(level, str)
            else level
        )

    match levels:
        case (level_spec,) if isinstance(level_spec, (str, int)):
            levels_spec = {"trickkiste": level_spec}
        case ((_logger, _level_spec), *_rest):
            assert isinstance(levels, tuple)
            levels_spec = {
                s[0] if isinstance(s[0], str) else s[0].name: s[1]
                for s in levels
                if isinstance(s, tuple)
            }
        case (others, (_logger, _level_spec), *_rest) if isinstance(
            others, (str, int)
        ):
            others_level = others
            levels_spec = {
                s[0] if isinstance(s[0], str) else s[0].name: s[1]
                for s in levels
                if isinstance(s, tuple)
            }
        case _:
            raise TypeError(f"unsupported arguments: {levels}")

    named_levels: dict[str | None, int] = {
        None: (
            logging.DEBUG
            if "ALL_DEBUG" in levels_spec.values()
            else level_of(others_level)
        ),
        "trickkiste": logging.DEBUG
        if "ALL_DEBUG" in levels_spec.values()
        else logging.INFO,
        **{
            logger_name: level_of(level_spec)
            for logger_name, level_spec in levels_spec.items()
        },
    }

    for handler in logging.getLogger().handlers:
        handler.setLevel(min(named_levels.values()))

    for name, level in named_levels.items():
        logging.getLogger(name).setLevel(level)
